//=====================================================================
//
//  File:       Distinct.cs
//  Summary:    This file contains the implementation of the Distinct transformation
//              it is heavily based on RemoveDuplicates sample from Microsoft, the code
//              ha been adapted and simplified to reach a very simple and memory safe
//              component in order to have the DISTINCT of some columns in a flow.    
//  Date:       07 february 2007
//  Web:        www.sqlbi.eu
//  Mail:       alberto.ferrari@sqlbi.eu
//
//===================================================================== 
using System;
using System.Collections;
using System.Data;
using Microsoft.SqlServer.Dts.Pipeline;
using Microsoft.SqlServer.Dts.Pipeline.Wrapper;
using Microsoft.SqlServer.Dts.Runtime;
using Microsoft.SqlServer.Dts.Runtime.Wrapper;
using System.Globalization;

namespace SqlBi.SSIS.Components {
    /// <summary>
    /// 
    /// This sample demonstrates a data flow transformation component with asynchronous outputs.
    /// 
    /// </summary>
    [DtsPipelineComponent (
        DisplayName = "Distinct",
        Description = "Finds and removes duplicate rows.",
        IconResource = "SqlBi.SSIS.Components.Distinct.ico",
        UITypeName = "SqlBi.SSIS.Components.DistinctUI, Distinct, Version=1.0.0.0, Culture=neutral,PublicKeyToken=729551dd7e47caaf",
        ComponentType = ComponentType.Transform)]
    public class Distinct : PipelineComponent {
        bool cancelEvent;

        Buffer distinctBufferCache;
        ColumnInfos colInfoDistinctOutput;

        /// <summary>
        /// 
        /// Runtime member. Used to cache the output buffer received in PrimeOutput. This buffer contains
        /// the rows in the output column collection of the meta data.
        /// 
        /// </summary>
        PipelineBuffer distinctBuffer;

        /// <summary>
        /// 
        /// Called when the component is initally added to a data flow task. 
        /// Create and configure the input and outputs of the component.
        /// 
        /// </summary>
        public override void ProvideComponentProperties () {
            ///
            /// Support resettability.
            /// 
            /// The base class calls RemoveAllInputsOutputsAndCustomProperties to reset the
            /// component. Used here to highlight the functionality of the base class.
            /// 
            base.RemoveAllInputsOutputsAndCustomProperties ();

            //
            //    Let the base component add the input and output.
            //
            base.ProvideComponentProperties ();

            //
            //    Name the input and output, and make the output asynchronous.
            //
            ComponentMetaData.InputCollection[0].Name = "Input";

            IDTSOutput90 distinctOutput = ComponentMetaData.OutputCollection[0];
            distinctOutput.Name = "DistinctOutput";
            distinctOutput.Description = "Distinct rows are directed to this output.";
            distinctOutput.SynchronousInputID = 0;
        }

        /// <summary>
        /// 
        /// Called repeatedly when the component is edited in the designer, and once at the beginning of execution.
        /// Verifies the following:
        /// 
        /// 1. The component has 2 outputs.
        /// 2. The component has 1 input.
        /// 3. There is an output column in each output for each input column.
        /// 4. The base class validation succeeds.
        /// 
        /// </summary>
        /// <returns></returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage ("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes"), CLSCompliant (false)]
        public override DTSValidationStatus Validate () {
            if (ComponentMetaData.OutputCollection.Count != 1) {
                InternalFireError (Properties.Resources.Exactly1Output);
                return DTSValidationStatus.VS_ISCORRUPT;
            }

            if (ComponentMetaData.InputCollection.Count != 1) {
                InternalFireError (Properties.Resources.MoreThan1Input);
                return DTSValidationStatus.VS_ISCORRUPT;
            }

            IDTSInput90 input = ComponentMetaData.InputCollection[0];
            IDTSOutput90 distinct = ComponentMetaData.OutputCollection[0];

            //
            // If there are more output columns than input columns, fail validation
            // by returning VS_NEEDSNEWMETADATA because ReintializeMetaData fixes this problem.
            //
            if (input.InputColumnCollection.Count != distinct.OutputColumnCollection.Count) {
                ComponentMetaData.FireError (0, ComponentMetaData.Name,
                    Properties.Resources.NumberInputColumnsNotMatchOutputColumns
                    + distinct.Name + Properties.Resources.Output, "", 0, out cancelEvent);
                return DTSValidationStatus.VS_NEEDSNEWMETADATA;
            }

            //
            // If there isn't a matching output column for each input column 
            // then the component meta data is out of sync. Return VS_NEEDSNEWMETADATA
            // because ReinitializeMetaData method fixes this problem.
            //
            foreach (IDTSInputColumn90 col in input.InputColumnCollection) {
                if (!OutputColumnExists (distinct, col.LineageID)) {
                    ComponentMetaData.FireError (0, ComponentMetaData.Name, "The input column " + col.Name + " is not found in the " + distinct.Name + " output.", "", 0, out cancelEvent);
                    return DTSValidationStatus.VS_NEEDSNEWMETADATA;
                }
            }

            //
            // Finally, call the base class, which validates that the LineageID of each column in the input collection
            // matches the LineageID of a column in the VirtualInputColumnCollection. 
            //
            return base.Validate ();
        }

        /// <summary>
        /// 
        /// The SetUsageType method is called when a column from the virtual input collection is selected for use by the component.
        /// The virtual input collection contains the collection of columns that the upstream component is providing to this component.
        /// If the column DataType is DT_IMAGE it is rejected, since BLOB datatypes are not supported by this component.
        /// 
        /// Based on the usageType parameter the column is either added, removed, or rejected by the component. 
        /// 
        ///     UsageType.UT_IGNORED    - The column was previously added to the input collection. The base class removes the 
        ///                               column from the input. The matching output columns are removed.
        ///     UsageType.UT_READONLY   - The column is added to the input collection of the component.
        ///     UsageType.UT_READWRITE  - Since this component does not modify columns this usagetype is rejected.
        /// 
        /// </summary>
        /// <param name="inputID">The ID of the input the column is mapped to.</param>
        /// <param name="virtualInput">The IDTSVirtualInput90 for the input.</param>
        /// <param name="lineageID">The lineageID of the virtual input column.</param>
        /// <param name="usageType">The usagetype of the column.</param>
        /// <returns>The newly created input column.</returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage ("Microsoft.Usage", "CA2201:DoNotRaiseReservedExceptionTypes"), CLSCompliant (false)]
        public override IDTSInputColumn90 SetUsageType (int inputID, IDTSVirtualInput90 virtualInput, int lineageID, DTSUsageType usageType) {
            if (virtualInput == null) {
                throw new ArgumentNullException ("virtualInput");
            }

            IDTSVirtualInputColumn90 vCol = virtualInput.VirtualInputColumnCollection.GetVirtualInputColumnByLineageID (lineageID);
            IDTSInputColumn90 col = null;

            //
            // No support for BLOB image columns.
            // 
            if (vCol.DataType == DataType.DT_IMAGE) {
                throw new Exception (Properties.Resources.BinaryDataTypesNotSupported);
            }

            //
            // If the usageType is UT_IGNORED, then the column is being removed.
            // So remove it from the outputs also.
            // 
            switch (usageType) {
                case DTSUsageType.UT_IGNORED :
                    col = base.SetUsageType (inputID, virtualInput, lineageID, usageType);
                    RemoveOutputColumn (ComponentMetaData.OutputCollection[0].ID, vCol.LineageID);
                    break;
                case DTSUsageType.UT_READONLY :
                    col = base.SetUsageType (inputID, virtualInput, lineageID, usageType);
                    AddOutputColumn (ComponentMetaData.OutputCollection[0].ID, col);
                    break;
                case DTSUsageType.UT_READWRITE:
                    throw new Exception (Properties.Resources.ReadWriteNotSupported);
            }

            return col;
        }

        /// <summary>
        /// 
        /// Called when VS_NEEDSNEWMETADATA is returned from Validate. 
        /// Reset all of the output columns.
        /// 
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage ("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes")]
        public override void ReinitializeMetaData () {
            IDTSOutput90 distinctOutput = ComponentMetaData.OutputCollection[0];
            AddInputColumnsToOutput (ComponentMetaData.OutputCollection[0].ID);
        }

        /// <summary>
        /// 
        /// Called when the path connecting the component to an upstream component is deleted. 
        /// Clear the columns from the output collections.
        /// 
        /// </summary>
        /// <param name="inputID"></param>
        public override void OnInputPathDetached (int inputID) {
            base.OnInputPathDetached (inputID);

            ComponentMetaData.OutputCollection[0].OutputColumnCollection.RemoveAll ();
        }

        /// <summary>
        /// 
        /// Called to add an input to the ComponentMetaData. This component doesn't allow adding inputs,
        /// so an exception is thrown.
        /// 
        /// </summary>
        /// <param name="insertPlacement"></param>
        /// <param name="inputID"></param>
        /// <returns></returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage ("Microsoft.Usage", "CA2201:DoNotRaiseReservedExceptionTypes"), CLSCompliant (false)]
        public override IDTSInput90 InsertInput (DTSInsertPlacement insertPlacement, int inputID) {
            throw new Exception (Properties.Resources.Component
                + ComponentMetaData.Name
                + Properties.Resources.NotAllowAddingInputs);
        }

        /// <summary>
        /// 
        /// Called to add an output to the ComponentMetaData. This component doesn't allow adding outputs,
        /// so an exception is thrown.
        /// 
        /// </summary>
        /// <param name="insertPlacement"></param>
        /// <param name="outputID"></param>
        /// <returns></returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage ("Microsoft.Usage", "CA2201:DoNotRaiseReservedExceptionTypes"), CLSCompliant (false)]
        public override IDTSOutput90 InsertOutput (DTSInsertPlacement insertPlacement, int outputID) {
            throw new Exception (Properties.Resources.Component
                + ComponentMetaData.Name
                + Properties.Resources.NotAllowAddingOutputs);
        }

        /// <summary>
        /// 
        /// Called to add an output to the ComponentMetaData. This component doesn't allow adding outputs,
        /// so an exception is thrown.
        /// 
        /// </summary>
        /// <param name="outputID"></param>
        /// <param name="outputColumnIndex"></param>
        /// <param name="name"></param>
        /// <param name="description"></param>
        /// <returns></returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage ("Microsoft.Usage", "CA2201:DoNotRaiseReservedExceptionTypes"), CLSCompliant (false)]
        public override IDTSOutputColumn90 InsertOutputColumnAt (int outputID, int outputColumnIndex, string name, string description) {
            throw new Exception (Properties.Resources.OutputColumnsCannotBeAdded
                + ComponentMetaData.Name);
        }

        /// <summary>
        /// 
        /// Called when the custom property on an output column is set. 
        /// Disallow modifications of the InputColumnLineageID custom property by 
        /// throwing an exception.
        /// 
        /// </summary>
        /// <param name="outputID"></param>
        /// <param name="outputColumnID"></param>
        /// <param name="propertyName"></param>
        /// <param name="propertyValue"></param>
        /// <returns></returns>
        [System.Diagnostics.CodeAnalysis.SuppressMessage ("Microsoft.Usage", "CA2201:DoNotRaiseReservedExceptionTypes"), CLSCompliant (false)]
        public override IDTSCustomProperty90 SetOutputColumnProperty (int outputID, int outputColumnID, string propertyName, object propertyValue) {
            if (propertyName == "InputColumnLineageID") {
                throw new Exception (Properties.Resources.InputColumnLineageIDPropertyReadOnly);
            }

            return base.SetOutputColumnProperty (outputID, outputColumnID, propertyName, propertyValue);
        }

        /// <summary>
        /// <summary>
        /// 
        /// For each column in the input collection, add a corresponding column to the 
        /// output column collection of the specified output object.
        /// 
        /// </summary>
        /// <param name="outputID">The ID of the output to add the columns to.</param>
        private void AddInputColumnsToOutput (int outputID) {
            IDTSInput90 input = ComponentMetaData.InputCollection[0];
            IDTSOutput90 output = ComponentMetaData.OutputCollection.GetObjectByID (outputID);

            output.OutputColumnCollection.RemoveAll ();

            foreach (IDTSInputColumn90 Col in input.InputColumnCollection)
                AddOutputColumn (outputID, Col);
        }


        /// <summary>
        /// 
        /// Creates an output column in the output identified by the OutputID parameter.
        /// Adds a custom property to the output containing the lineageID of the input column. This property is
        /// used during Validate, to verify that each input column has a matching output column.
        /// 
        /// </summary>
        /// <param name="OutputID">The ID of the output to create the column in.</param>
        /// <param name="inputColumn">The input column used to set the properties of the output column.</param>
        private void AddOutputColumn (int OutputID, IDTSInputColumn90 inputColumn) {
            IDTSOutput90 output = ComponentMetaData.OutputCollection.GetObjectByID (OutputID);
            IDTSOutputColumn90 outputColumn = output.OutputColumnCollection.New ();
            outputColumn.Name = inputColumn.Name;

            //
            //    Store the lineageID of the input column in a custom property of the output column.
            //
            IDTSCustomProperty90 inputColLineageID = outputColumn.CustomPropertyCollection.New ();
            inputColLineageID.Name = "InputColumnLineageID";
            inputColLineageID.Value = inputColumn.LineageID;

            outputColumn.SetDataTypeProperties (inputColumn.DataType, inputColumn.Length, inputColumn.Precision, inputColumn.Scale, inputColumn.CodePage);
        }

        /// <summary>
        /// 
        /// Removes the output column that references the input column. The reference is maintained in the 
        /// InputColumnLineageID custom property. 
        /// 
        /// </summary>
        /// <param name="OutputID">The ID of the output.</param>
        /// <param name="inputColumn">The lineageID of the input column.</param>
        private void RemoveOutputColumn (int OutputID, int LineageID) {
            IDTSOutput90 output = ComponentMetaData.OutputCollection.GetObjectByID (OutputID);
            foreach (IDTSOutputColumn90 col in output.OutputColumnCollection) {
                IDTSCustomProperty90 inputColumnLineageID = col.CustomPropertyCollection["InputColumnLineageID"];
                if (inputColumnLineageID.Value != null && (int) inputColumnLineageID.Value == LineageID) {
                    output.OutputColumnCollection.RemoveObjectByID (col.ID);
                    break;
                }
            }
        }

        /// <summary>
        /// 
        /// Searches the output column collection for a column containing the provided lineageID
        /// in the InputColumnLineageID custom property.
        /// 
        /// </summary>
        /// <param name="LineageID">The lineageID of the input column.</param>
        /// <param name="output">The IDTSOutput to search.</param>
        /// <returns>true if found; otherwise false.</returns>
        private static bool OutputColumnExists (IDTSOutput90 output, int LineageID) {
            foreach (IDTSOutputColumn90 outColumn in output.OutputColumnCollection) {
                //
                // If the custom property doesn't exist, return false.
                //
                if (outColumn.CustomPropertyCollection.Count != 1)
                    return false;

                if ((int) outColumn.CustomPropertyCollection[0].Value == LineageID)
                    return true;
            }
            return false;
        }

        /// 
        /// Called prior to PrimeOutput and ProcessInput. 
        /// Creates the internal buffer and stores the buffer column indexs of the input and output columns
        /// 
        /// </summary>
        public override void PreExecute () {
            colInfoDistinctOutput = new ColumnInfos ();
            colInfoDistinctOutput.AddColumnInformation (BufferManager, ComponentMetaData.OutputCollection[0], ComponentMetaData.InputCollection[0]);

            distinctBufferCache = new Buffer (colInfoDistinctOutput);
        }

        /// <summary>
        /// 
        /// This method is called repeatedly during package execution. It is called each time the data flow task
        /// has a full buffer provided by an upstream component. 
        /// 
        /// We collect each chunk of data in inputBufferCache and then we merge it with the actual distinctBufferCache
        /// that holds the sorted collection of distinct rows read so far.
        /// 
        /// </summary>
        /// <param name="inputID">The ID of the input object of the component.</param>
        /// <param name="buffer">The input buffer.</param>
        public override void ProcessInput (int inputID, PipelineBuffer buffer) {
            if (distinctBuffer != null) {
                if (!buffer.EndOfRowset) {
                    while (buffer.NextRow ()) {
                        Row row = distinctBufferCache.NewRow (buffer);
                        if (distinctBufferCache.AddRow (row)) {
                            AddRowToOutputBuffer (distinctBuffer, row, colInfoDistinctOutput);
                        }
                    }
                } else {
                    distinctBuffer.SetEndOfRowset ();
                }
            }
        }

        /// <summary>
        /// 
        /// Called after PreExecute. The data flow task provides a buffer for each output 
        /// that is connected to a downstream component. Check the location of each output specified in the
        /// outputIDs array, and based on the index of the output cache the output buffer to the appropriate
        /// private member variable.
        /// 
        /// </summary>
        /// <param name="outputs">The number of outputs.</param>
        /// <param name="outputIDs">An array of output ids.</param>
        /// <param name="buffers">An array of output buffers.</param>
        public override void PrimeOutput (int outputs, int[] outputIDs, PipelineBuffer[] buffers) {
            distinctBuffer = buffers[0];
        }

        private void InternalFireError (string Message) {
            ComponentMetaData.FireError (0, ComponentMetaData.Name, Message, "", 0, out cancelEvent);
        }

        /// <summary>
        /// 
        /// Adds the columns in the provided row object to the provided output buffer.
        /// 
        /// </summary>
        /// <param name="buffer">The output buffer the rows are added to.</param>
        /// <param name="row">The row object containing the columns that are added to the buffer.</param>
        /// <param name="colInfos">A collection of ColumnInfo objects. There is a one
        /// to one mapping between the ColumnInfo objects and the columns in the row.</param>
        private static void AddRowToOutputBuffer (PipelineBuffer buffer, Row row, ColumnInfos colInfos) {
            buffer.AddRow ();

            for (int col = 0; col < colInfos.Count; col++)
                buffer[colInfos[col].OutputBufferColumnIndex] = row[col];
        }
   }
}