using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;

namespace SqlBi.SSIS.Components {
    public partial class DistinctForm : Form {
        internal event GetAvailableColumnsEventHandler GetAvailableColumns;
        internal event ChangeInputOutputColumnsEventHandler SetInputOutputColumns;
        internal event ChangeInputOutputColumnsEventHandler DeleteInputOutputColumns;

        // This flag is set to true while loading the state from the component, to disable
        // the grid events during that time.
        bool isLoading;

        public DistinctForm () {
            InitializeComponent ();
        }

        /// <summary>
        /// 
        /// Hooking up a data flow element to a grid cell.
        /// 
        /// </summary>
        /// <param name="cell"></param>
        /// <param name="dataFlowElement"></param>
        private static void SetGridCellData (DataGridViewCell cell, DataFlowElement dataFlowElement) {
            cell.Value = dataFlowElement.ToString ();
            cell.Tag = dataFlowElement.Tag;
            cell.ToolTipText = dataFlowElement.ToolTip;
        }

        /// <summary>
        /// 
        /// Loading available columns to the upper grid.
        /// 
        /// </summary>
        [System.Diagnostics.CodeAnalysis.SuppressMessage ("Microsoft.Performance", "CA1817:DoNotCallPropertiesThatCloneValuesInLoops")]
        private void LoadAvailableColumns () {
            if (this.GetAvailableColumns != null) {
                AvailableColumnsArgs args = new AvailableColumnsArgs ();
                IAsyncResult res = this.GetAvailableColumns.BeginInvoke (this, args, null, null);
                this.GetAvailableColumns.EndInvoke (res);

                if (args.AvailableColumns != null && args.AvailableColumns.Length > 0) {
                    this.dataGridViewAvailableColumns.Rows.Add (args.AvailableColumns.Length);
                    for (int i = 0; i < args.AvailableColumns.Length; i++) {
                        AvailableColumnElement availableColumnRow = args.AvailableColumns[i];

                        //
                        // Filling the cells.
                        //
                        this.dataGridViewAvailableColumns.Rows[i].Cells[this.gridColumnCheckbox.Index].Value = availableColumnRow.Selected;
                        SetGridCellData (this.dataGridViewAvailableColumns.Rows[i].Cells[this.gridColumnAvailableColumns.Index], availableColumnRow.AvailableColumn);
                    }
                }
            }
        }

        /// <summary>
        /// 
        /// Adding newly selected columns to the Data Flow and the bottom grid.
        /// 
        /// </summary>
        /// <param name="args"></param>
        private void SetColumns (SetInputOutputColumnsArgs args) {
            if (this.SetInputOutputColumns != null) {
                IAsyncResult res = this.SetInputOutputColumns.BeginInvoke (this, args, null, null);
                this.SetInputOutputColumns.EndInvoke (res);
            }
        }

        /// <summary>
        /// 
        /// Deleting unselected columns from the Data Flow and the bottom grid.
        /// 
        /// </summary>
        /// <param name="args"></param>
        private void DeleteColumns (SetInputOutputColumnsArgs args) {
            if (this.DeleteInputOutputColumns != null) {
                IAsyncResult res = this.DeleteInputOutputColumns.BeginInvoke (this, args, null, null);
                this.DeleteInputOutputColumns.EndInvoke (res);

            }
        }

        //
        // Loading available and previously selected columns.
        //
        protected override void OnLoad (EventArgs e) {
            base.OnLoad (e);

            this.isLoading = true;
            try {
                this.LoadAvailableColumns ();
            } finally {
                this.isLoading = false;
            }
        }

        //
        // Commiting change in the checbox columns so it will trigger CellValueChanged immediately.
        //
        private void dataGridViewAvailableColumns_CurrentCellDirtyStateChanged (object sender, EventArgs e) {
            if (this.dataGridViewAvailableColumns.CurrentCell != null && this.dataGridViewAvailableColumns.CurrentCell is DataGridViewCheckBoxCell) {
                this.dataGridViewAvailableColumns.CommitEdit (DataGridViewDataErrorContexts.Commit);
            }
        }

        private void dataGridViewAvailableColumns_CellValueChanged (object sender, DataGridViewCellEventArgs e) {
            //
            // Ignoring this event while loading columns.
            //
            if (this.isLoading) {
                return;
            }

            if (e.ColumnIndex == this.gridColumnCheckbox.Index && e.RowIndex >= 0) {
                DataGridViewCheckBoxCell checkBoxCell = this.dataGridViewAvailableColumns.CurrentCell as DataGridViewCheckBoxCell;
                DataGridViewCell columnCell = this.dataGridViewAvailableColumns.Rows[e.RowIndex].Cells[this.gridColumnAvailableColumns.Index];

                SetInputOutputColumnsArgs args = new SetInputOutputColumnsArgs ();
                args.VirtualColumn = new DataFlowElement (columnCell.Value.ToString (), columnCell.Tag);

                if ((bool) checkBoxCell.Value) {
                    // Set used columns if the checkbox is selected.
                    this.SetColumns (args);
                } else {
                    // Delete columns if the checkbox is unselected.
                    this.DeleteColumns (args);
                }
            }
        }

    }

    internal struct AvailableColumnElement {
        public bool Selected;
        public DataFlowElement AvailableColumn;
    }

    internal struct SelectedInputOutputColumns {
        public DataFlowElement InputColumn;
        public DataFlowElement DistinctColumn;
        public DataFlowElement DuplicateColumn;
    }

    internal class AvailableColumnsArgs {
        public AvailableColumnElement[] AvailableColumns;
    }

    internal delegate void GetAvailableColumnsEventHandler (object sender, AvailableColumnsArgs args);

    internal class SetInputOutputColumnsArgs {
        public DataFlowElement VirtualColumn;
        public SelectedInputOutputColumns GeneratedColumns;
        public bool CancelAction;
    }

    internal delegate void ChangeInputOutputColumnsEventHandler (object sender, SetInputOutputColumnsArgs args);

}